;;;;;;;;;;;;;;;;;;;;;;;;;;; -*- Mode: Emacs-Lisp -*- ;;;;;;;;;;;;;;;;;;;;;;;;;;
;; pvs-ilisp-mods.el -- Functions originally defined in ILISP, but modified
;;                      for use in PVS
;; Author          : Sam Owre
;; Created On      : Fri Mar  4 16:46:49 1994
;; Last Modified By: Sam Owre
;; Last Modified On: Tue Jun  4 13:21:35 1996
;; Update Count    : 14
;; Status          : Unknown, Use with caution!
;; 
;; HISTORY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; This is modified primarily to stop from popping to the *pvs* buffer.
;;; Originally in <ILISP>/ilisp-mod.el

(defun ilisp (name setup)
  "Run an inferior LISP process NAME, input and output via buffer *name*.
If there is a process already running in *name*, just switch to that buffer.
Takes the program name from the variable ilisp-program.
\(Type \\[describe-mode] in the process buffer for a list of commands.)"
  (set-buffer ilisp-buffer)
  (if (not (comint-check-proc ilisp-buffer))
      (let* ((dialect (car ilisp-dialect))
	     (program ilisp-program)
	     (args (lisp-command-args program))
	     ;; Use pipes so that strings can be long
	     (process-connection-type nil)
	     (names (format "%s" name))
	     start)
	(apply 'make-comint name (car args) nil (cdr args))
	(comint-setup-ipc)
	;; Because comint-mode kills all buffer-local variables in
	;; fsf-19 we have to re-call the setup here.
	(funcall setup name)
	(setq major-mode 'ilisp-mode
	      mode-name "ILISP")
	(use-local-map ilisp-mode-map) ;; Added by SO
	(rplaca (car comint-send-queue) 
		(function (lambda ()
			    (run-hooks 'ilisp-init-hook))))
	(setq ilisp-initialized (lisp-del ilisp-buffer ilisp-initialized))
	(if (not (lisp-memk names ilisp-buffers 'car))
	    (setq ilisp-buffers (cons (list names) ilisp-buffers)))
	;;(lisp-pop-to-buffer ilisp-buffer) ;; Commented out by SO
	(setq start (window-start (selected-window))
	      ilisp-program program)
	(goto-char (point-max))
	(insert (format "Starting %s ...\n" ilisp-program))
	(set-marker (process-mark (ilisp-process)) (point))
	(funcall comint-update-status 'start)
	(if ilisp-motd
	    (progn (lisp-display-output (format ilisp-motd ilisp-version))
		   (set-window-start (selected-window) start)))
	(if (not ilisp-prefix-match) (require 'completer)))
      ;;(lisp-pop-to-buffer ilisp-buffer) ;; Commented out by SO
      )
  (use-local-map ilisp-use-map)
  ;; This is necessary to get mode documentation to come out right
  (set-default 'ilisp-use-map ilisp-use-map))

;;; Modified comint-log to restrict the size of the log buffer, and remove
;;; garbage.  Originally defined in <ILISP>/comint-ipc.el

(defvar comint-log-size 15000)
(defvar comint-log-verbose t)
(defvar comint-noise-regexp
  "> \\|(ILISP:ilisp-restore)\\|NIL\\|^[\n\t ]*$")

(defun comint-log (process string &optional output)
  "Log to PROCESS, STRING marking as optional OUTPUT."
  (if (and comint-log
	   (or comint-log-verbose
	       (not (string-match comint-noise-regexp string))))
      (save-excursion
	(set-buffer (get-buffer-create (process-name process)))
	(goto-char (point-max))
	(if output
	    (progn
	      (insert "{") (insert string) (insert "}"))
	    (insert string))
	(if (> (buffer-size) comint-log-size)
	    (delete-region (point-min) (- (buffer-size) comint-log-size))))))


;;; Modified to always set the buffer to comint-original-buffer
;;; Original in <ILISP>/comint-ipc.el

(defun comint-process-filter (process output)
  "Filter PROCESS OUTPUT.  See comint-send for more information.  The
first element of the comint-send-queue is the current send entry.  If
the entry has a nil no-insert flag, insert the results into the
process buffer.

If the send is an interrupt, comint-interrupt-start is funcalled on
the output and should return the start of the output of an interrupt.

comint-prompt-status is called with the old prompt and the last line.
It should return 'error if the last line is an error, T if it is a
prompt and nil otherwise.  It should also update the process status by
funcalling comint-update-status.

If there is a send handler, it is called with \(error-p wait-p message
output prompt) and should determine what sort of notification is
appropriate and return T if errors should be fixed and NIL otherwise.

If the prompt is an error, then comint-fix-error will be sent to fix
the error. 

When there is a prompt in the output stream, the next send will be
dispatched unless the wait flag for the send is a string.  If it is a
string, then results will be discarded until one matches the string as
a regexp.

Output to the process should only be done through the functions
comint-send or comint-default-send, or results will be mixed up."
  (let* ((inhibit-quit t)
	 (window (selected-window))
	 (comint-original-buffer (prog1 (current-buffer)
				   (set-buffer (process-buffer process))))
	 (match-data (match-data))
	 (send (car comint-send-queue))
	 (no-insert (cdr send))
	 (wait-p (cdr no-insert))
	 (messagep (cdr (cdr wait-p)))
	 (handler (cdr messagep))
	 (running (cdr handler))
	 (old-prompt (cdr running))
	 (line (cdr old-prompt))
	 (result (car (cdr line)))
	 (old-result (car result))
	 (no-insert (car no-insert))
	 (message (car messagep))
	 (wait-p (car wait-p))
	 (sync (stringp wait-p)))
    (comint-log process output t)
    ;; Remove leading whitespace
    (if (and (null old-result)
	     (save-excursion (goto-char (process-mark process)) (bolp))
	     (eq (string-match "[ \t]*\n" output) 0))
	(setq output (substring output (match-end 0))))
    (rplaca result (concat old-result output))
    (while (string-match "\n" (car result) (car line))
      (rplaca line (match-end 0)))
    (if (not (or sync no-insert))
	(progn
	  (comint-insert output)
	  ;; Throw away output if storing in buffer
	  (rplaca result (substring (car result) (car line)))
	  (rplaca line 0)))
    (if (consp (car running))		;Waiting for interrupt
	(let ((split (funcall comint-interrupt-start (car result))))
	  (if split
	      (let ((interrupted (car running)))
		;; Store output to previous send
		(rplaca (comint-send-variables interrupted) 
			(substring (car result) 0 split))
		(rplaca result (substring (car result) (car line)))
		(rplaca line 0)
		(rplaca running t)))))
    (if (not (consp (car running)))	;Look for prompt
	(let* ((last (substring (car result) (car line)))
	       (is-prompt
		(funcall comint-prompt-status (car old-prompt) last)))
	  (if is-prompt
	      (let* ((output
		      (if (or no-insert sync)
			  (funcall comint-output-filter 
				   (substring (car result) 0 (car line)))))
		     (handler (car handler))
		     (error (eq is-prompt 'error)))
		(setq old-result (car result))
		(rplaca result output)
		(rplacd result (if error (list last) last))
		(setq comint-output (car result)
		      comint-errorp 
		      (or error
			  (and comint-error-regexp
			       comint-output
			       (string-match comint-error-regexp
					     comint-output))))
		(unwind-protect
		     (if handler
			 (setq handler
			       (funcall handler comint-errorp wait-p
					message output last)))
		  (if (and error handler no-insert comint-fix-error)
		      (setq comint-send-queue 
			    (cons (list comint-fix-error t nil 'fix
					"Fixing error" nil
					nil nil 0 (cons nil nil))
				  ;; We may have aborted
				  (or (cdr comint-send-queue)
				      comint-send-queue))))
		  (if sync
		      (let ((match (string-match wait-p old-result)))
			(if match
			    (progn
			      (rplaca
			       (cdr (cdr (cdr (cdr (car comint-end-queue)))))
			       "Done")
			      (if (not no-insert)
				  (comint-insert 
				   (concat 
				    (substring old-result 0 match)
				    (substring old-result (match-end 0)))))
			      (rplaca result (substring old-result
							match (car line)))
			      (rplaca messagep "Done")
			      (rplaca running nil)
			      (comint-dispatch-send process))))
		      ;; Not waiting
		      (rplaca messagep "Done")
		      (rplaca running nil)
		      (comint-dispatch-send process))))
	      (rplacd result nil))))
    (store-match-data match-data)
    (if (or t ;; Added by SO
	    (get-buffer-window comint-original-buffer)
	    (eq (window-buffer (minibuffer-window)) comint-original-buffer))
	(set-buffer comint-original-buffer))))

(defun comint-send (process string 
			    &optional 
			    no-insert
			    wait
			    status 
			    message
			    handler
			    after)
  "Do a send to PROCESS of STRING.  Optionally specify NO-INSERT,
WAIT, STATUS, MESSAGE, HANDLER and AFTER.  Without optional arguments,
this is just like process-send-string.  If STRING is not a string,
then it is assumed to be an elisp function and will be called when
encountered in the send queue.  The send will be the next one if WAIT,
after the last send if AFTER, otherwise it will be put at the end of
the queue. If WAIT is non-NIL or on the first send to a busy inferior,
the inferior will be interrupted if possible, see comint-interrupt for
more information.  Once the send is sent, the process status will be
STATUS or 'run.  Output of the send will be inserted into the process
buffer unless NO-INSERT.  This function returns a list of \(result .
prompt).  If WAIT is a string, output will be inserted until one
matches the string as a regexp.  If WAIT is T, then PROMPT will have
the prompt when finished and RESULT will have the output.  If PROMPT
is a list, then there was an error.  If WAIT is not T, then the list
returned will change when the send has been sent and is finished.  If
HANDLER is nil it will be set to comint-handler.  If it is T, errors
will be ignored.  When a send is finished, it calls handler with
\(error-p WAIT MESSAGE output prompt) which decides what to do with
the output.

VARIABLES:

comint-always-scroll will cause all process output to be visible.

comint-fix-error is the string used to fix errors.

comint-continue is the string used to continue after an interrupt.

comint-interrupt-regexp is the default regexp to use in finding the
start of the interrupt text.  

comint-error-regexp will set comint-errorp if found in the process output.  

FUNCTIONS:  Each of the functions in these variables is called with
the buffer set to the appropriate process buffer and
comint-original-buffer bound to the buffer current when the process
filter was called.  

comint-update-status is a function \(status) that is called each time
the process status changes.

comint-prompt-status is called with the old prompt and the last line.
It should return 'error if the last line is an error, T if it is a
prompt and nil otherwise.  It should also update the process status by
funcalling comint-update-status.

comint-output-filter is a function \(output) for sends with NO-INSERT.
It should return the output string.

comint-interrupt-start is a function \(output) that returns the start
of the interrupt text in output using comint-interrupt-regexp to find it."
  (save-excursion
    (set-buffer (process-buffer process))
    (let* ((inhibit-quit t)
	   (send (list string 
		       no-insert
		       wait
		       (or status 'run)
		       message 
		       (if (eq handler t) nil (or handler comint-handler))
		       ;; running, old-prompt, line
		       nil nil 0
		       ;; (output . prompt)
		       (cons nil nil)))
	   (pointer (comint-send-results send))
	   (top-level (eq comint-send-queue comint-end-queue))
	   (end (car comint-end-queue))
	   (current (car comint-send-queue))
	   (prompt (cdr (comint-send-results current)))
	   (ok nil))
      (setq comint-aborting nil)
      (if (and top-level
	       (equal comint-status " :ready")
	       (or (stringp wait) prompt comint-queue-emptied))
	  (progn
	    (setq comint-send-queue (cons send comint-send-queue))
	    (comint-dispatch-send process))
	  (if (or (and wait (not after) (not prompt)) top-level)
	      (comint-interrupt process send)
	      (let ((looking t) 
		    (next comint-send-queue))
		(if after
		    (while (and looking next)
		      (if (eq (car next) comint-last-send)
			  (progn
			    (rplacd next (cons send (cdr next)))
			    (setq looking nil)))
		      (setq next (cdr next))))
		(if looking
		    (progn
		      (rplaca comint-end-queue send)
		      (setq comint-end-queue
			    (rplacd comint-end-queue (cons end nil))))))))
      (setq comint-last-send send)
      (unwind-protect
	   (let ((inhibit-quit nil))
	     (if (eq wait t)
		 (while (not (cdr pointer))
		   (accept-process-output)
		   (sit-for 0)))
	     (setq ok pointer))
	(if (not ok)
	    (if (eq send (car comint-send-queue))
		(let ((interrupt 
		       ;; string no-insert wait status message handler
		       (list nil t nil 'interrupt "Interrupt" nil
			     ;; running old-prompt line (output . prompt)
			     send (car (cdr (comint-send-variables send)))
			     nil (cons nil nil)))) 
		  (setq comint-send-queue
			(cons interrupt (cdr comint-send-queue)))
		  (comint-interrupt-subjob))
		(setq comint-send-queue (delq send comint-send-queue))))))))


;;; Changed NIL to 'dispatch
;;; Original in <ILISP>/ilisp-snd.el

(defun ilisp-send (string &optional message status and-go handler)
  "Send STRING to the ILISP buffer, print MESSAGE set STATUS and
return the result if AND-GO is NIL, otherwise switch to ilisp if
and-go is T and show message and results.  If AND-GO is 'dispatch,
then the command will be executed without waiting for results.  If
AND-GO is 'call, then a call will be generated. If this is the first
time an ilisp command has been executed, the lisp will also be
initialized from the files in ilisp-load-inits.  If there is an error,
comint-errorp will be T and it will be handled by HANDLER."
  (ilisp-init t)
  (let ((process (ilisp-process))
	(dispatch (eq and-go 'dispatch)))
    (if message
	(message "%s" (if dispatch
			  (concat "Started " message)
			  message)))
    ;; No completion table
    (setq ilisp-original nil)
    (if (memq and-go '(t call))
	(progn (comint-send process string nil nil status message handler)
	       (if (eq and-go 'call)
		   (call-defun-lisp nil)
		   (switch-to-lisp t t))
	       nil)
	(let* ((save (ilisp-value 'ilisp-save-command t))
	       (result
		(comint-send 
		 process
		 (if save (format save string) string)
		 ;; Interrupt without waiting
		 t (if (not dispatch) 'wait) status message handler)))
	  (if save 
	      (comint-send
	       process
	       (ilisp-value 'ilisp-restore-command t)
	       t 'dispatch 'restore "Restore" t t))  ;; nil -> 'dispatch  - SO
	  (if (not dispatch)
	      (progn
		(while (not (cdr result))
		  ;;(sit-for 0)
		  (accept-process-output process))
		(comint-remove-whitespace (car result))))))))

;;; Modified to not invoke message if noninteractive and not verbose

(defun ilisp-init (&optional waitp forcep sync)
  "Initialize the current inferior LISP if necessary by loading the
files in ilisp-load-inits.  Optional WAITP waits for initialization to
finish.  When called interactively, force reinitialization.  With a
prefix, get the binary extensions again."  
  (interactive 
   (list (if current-prefix-arg
	     (progn
	       (set-ilisp-value 'ilisp-init-binary-extension nil)
	       (set-ilisp-value 'ilisp-binary-extension nil)
	       nil))
	 t))
  (if (or forcep (not (ilisp-initialized)))
      (progn
	(unless (and noninteractive (= pvs-verbose 0))
	  (message "Started initializing ILISP"))
	(if (not ilisp-directory)
	    (setq ilisp-directory (or (ilisp-directory "ilisp.elc" load-path)
				      (ilisp-directory "ilisp.el" load-path))))
	(if (not (ilisp-value 'ilisp-initializing t))
	    (ilisp-init-internal sync))
	(if waitp
	    (while (ilisp-value 'ilisp-initializing t)
	      (accept-process-output)
	      (sit-for 0))))))

(defun ilisp-done-init ()
  "Make sure that initialization is done and if not dispatch another check."
  (if ilisp-load-files
      (comint-send-code (get-buffer-process (current-buffer))
			'ilisp-done-init)
      (if ilisp-initializing
	  (progn
	    (unless (and noninteractive (= pvs-verbose 0))	    
	      (message "Finished initializing %s" (car ilisp-dialect)))
	    (setq ilisp-initializing nil
		  ilisp-initialized
		  (cons (buffer-name (current-buffer)) ilisp-initialized))))))


;;; Modified to change the size of the popper window
;;; Original in <ILISP>/popper.el

(defun popper-bury-output (&optional no-error)
  "Bury the popper output signalling an error if not there unless
optional NO-ERROR is T."
  (interactive)
  (if (boundp 'epoch::version)
      (if (epoch::screen-mapped-p (popper-screen))
	  (epoch::iconify-screen (popper-screen)))
      (let ((buffer (popper-output-buffer)))
	(if buffer
	    (let* ((old (current-buffer))
		   (old-point (point))
		   (old-window (selected-window))
		   (output (get-buffer-window buffer))
		   (start (window-height output))
		   (parent (next-window output 'no))
		   (parent-point nil)
		   (height (window-height output))
		   (heights (popper-window-heights output)))
	      (bury-buffer buffer)
	      (delete-window output)
	      (popper-select parent)
	      (setq parent-point (point))
	      (while heights
		(unless (one-window-p)
		  (condition-case nil
		      (enlarge-window (- (+ height (car heights))
					 (window-height)))
		    (error nil)))
		(if start
		    (condition-case () (scroll-down start) (error nil)))
		(select-window (next-window (selected-window) 'no))
		(setq height 0 start nil)
		(setq heights (cdr heights)))
	      (goto-char parent-point)
	      (set-buffer old)
	      (unless (eq old-window output)
		(select-window old-window)
		(goto-char old-point)))
	    (if (not no-error) (popper-show-output))))))

;; modified to handle incompatible change in 19.23 (the interface
;; to where-is-internal changed)
;; original in <ILISP>/popper.el

(defun popper-show-output (&optional buffer size)
  "Bring the output window up showing optional BUFFER in window of
SIZE.  If SIZE is not specified, then shrink the window.  Finally
select the original window."
  (let* ((window (selected-window))
	 (old-buffer (window-buffer window))
	 (buffer (get-buffer-create
		  (or buffer (popper-first-buffer) 
		      (error "No popper buffers"))))
	 start parent
	 (min-height (+ window-min-height (or size window-min-height)))
	 (text-lines (save-excursion
		       (set-buffer buffer)
		       (count-lines (point-min) (point-max)))))
    (if (and (= text-lines 1) popper-use-message-buffer)
	(message 
	 (save-excursion
	   (set-buffer buffer) (buffer-substring (point-min) (point-max))))
    (setq popper-last-output-window window)
    (if (eq buffer old-buffer)
	(popper-shrink-window)
	(if (eq window (minibuffer-window)) 
	    (let* ((parent (popper-parent)))
	      (popper-bury-output t)
	      (select-window (setq window (or parent (previous-window)))))
	    (if (not (eq old-buffer (popper-output-buffer)))
		(popper-bury-output t)))
	(if (< (window-height window) min-height)
	    (enlarge-window (- min-height (window-height window))))
	(setq start (window-start window))
	(split-window nil size)
	(set-window-buffer window buffer)
	(set-buffer buffer)
	(setq popper-buffer t)
	(or popper-mode-line-text
	    (setq popper-mode-line-text
		  (list (format 
			 " %s bury, %s scroll" 
			 (key-description
			  (if (and (eq emacs-system 'emacs19)
				   (or (not (boundp 'emacs-minor-version))
				       (< emacs-minor-version 23)))
			      (where-is-internal
			       'popper-bury-output nil nil t)
			    (where-is-internal 'popper-bury-output nil t)))
			 (key-description 
			  (if (and (eq emacs-system 'emacs19)
				   (or (not (boundp 'emacs-minor-version))
				       (< emacs-minor-version 23)))
			      (where-is-internal 
			       'popper-scroll-output nil nil t)
			    (where-is-internal 'popper-scroll-output 
					       nil t)))))))
	(setq popper-output-buffers
	      (cons buffer (delq buffer popper-output-buffers)))
	(if (not size) (popper-shrink-window))
	(setq parent (next-window window 'no))
	(popper-select parent)
	;; Move the window so that top lines get covered unless it would
	;; cover point in which case point is at top of window
	(save-excursion
	  (set-window-start parent start)
	  (move-to-window-line (window-height window))
	  (set-window-start parent (point)))
	(let ((point (save-excursion (beginning-of-line) (point))))
	  (if (not (pos-visible-in-window-p point))
	      (set-window-start (selected-window) point)))
	(if (eq popper-last-output-window (minibuffer-window))
	    (select-window (minibuffer-window)))
	(set-buffer old-buffer)))))


;;; This is return-ilisp modified to put an extra backslash in front of any
;;; given ones.  Put
;;;  (define-key ilisp-mode-map "\C-m" 'return-ilisp-pvs)
;;; in your .pvsemacs file.

(defun return-ilisp-pvs ()
  "Grab the current expression with comint-get-old-input.  If we have
a complete sexp, send it.  Otherwise, indent appropriately."
  (interactive)
  (let ((proc (get-buffer-process (current-buffer))))
    (if (not proc)
	(error "Current buffer has no process")
	(let* ((pmark (process-mark proc))
	       (input1 (ilisp-get-old-input))
	       (input (protect-backslash-for-lisp input1)))
	  (if input
	      (let ((input-ring (ilisp-get-input-ring)))
		(if (>= (point) pmark)
		    (goto-char (point-max))
		    (goto-char pmark)
		    (insert input))
		(if (not ilisp-no-newline) (insert ?\n))
		(if (and (funcall comint-input-filter input)
			 (or (ring-empty-p input-ring)
			     (not (string= (ring-ref input-ring 0) input))))
		    (ilisp-ring-insert input-ring input))
		(funcall comint-input-sentinel input)
		;; Ugh, comint changing under my feet....
		(if (eq ilisp-emacs-version-id 'fsf-19)
		    (setq comint-input-ring-index nil))
		;; Nuke symbol table
		(setq ilisp-original nil)
		(funcall comint-input-sender proc input)
		(set-marker (process-mark proc) (point))
		(set-marker comint-last-input-end (point))
		(goto-char (point-max)))
	      (if (= pmark (point-max)) 
		  (let ((comint-send-newline t))
		    (if (not ilisp-no-newline) (insert ?\n))
		    (set-marker (process-mark proc) (point))
		    (funcall comint-input-sender proc ""))
		  (insert ?\n)
		  (save-restriction
		    (narrow-to-region pmark (point-max))
		    (funcall indent-line-function))))))))

(defun protect-backslash-for-lisp (string)
  (let ((pos (- (length string) 1))
	(chars nil))
    (while (>= pos 0)
      (let ((ch (aref string pos)))
	(when (and (= ch ?\\)
		   ;; \/ but not \\/
		   (or (and (= (car chars) ?/)
			    (or (= pos 0)
				(/= (aref string (1- pos)) ?\\)))
		       ;; /\ but not /\\
		       (and (/= (car chars) ?\\)
			    (> pos 0)
			    (= (aref string (1- pos)) ?/))))
	  (push ch chars))
	(push ch chars)
	(decf pos)))
    (concat chars)))

(when (memq emacs-system '(lemacs emacs20 emacs19 xemacs))
  (defkey-ilisp "\C-\\s" 'comint-previous-matching-input-from-input)
  (defkey-ilisp "\M-s" 'comint-previous-matching-input-from-input)

  (add-hook 'comint-mode-hook
    '(lambda ()
       (defun comint-previous-input (arg)
	 "Search backwards through input history for match for REGEXP.
\(Previous history elements are earlier commands.)
With prefix argument N, search for Nth previous match.
If N is negative, find the next or Nth next match.
Modified from original to not delete earlier command if some other command
intervenes."
	 (interactive "*p")
	 (setq arg (comint-search-arg arg))
	 (unless (memq last-command
		       '(comint-previous-input
			 comint-next-input))
	   (set-marker comint-last-input-start (point)))
	 (let ((pos (comint-previous-matching-input-string-position "." arg)))
	   ;; Has a match been found?
	   (if (null pos)
	       (error "Not found")
	       (setq comint-input-ring-index pos)
	       (message "History item: %d" (1+ pos))
	       (delete-region 
		;; Can't use kill-region as it sets this-command
		comint-last-input-start (point))
	       (insert (ring-ref comint-input-ring pos)))))))
  )


;;; Changed comint-sender to directly include the newline if it is needed,
;;; rather than sending the string and the newline in two separate calls.

(defun comint-sender (process string)
  "Send to PROCESS STRING with newline if comint-send-newline."
  (if comint-send-newline
      (comint-send-string process (concat string "\n"))
      (comint-send-string process string)))

;;; The following functions fix a problem with completion.el
;;; distributed with Emacs prior to 19.31

(when (and (eq emacs-system 'emacs19)
	   (boundp 'emacs-major-version)
	   (< emacs-minor-version 31))
(defun pvs-symbol-under-point ()
  "Returns the symbol that the point is currently on.
But only if it is longer than `completion-min-length'."
  (setq cmpl-saved-syntax (syntax-table))
  (set-syntax-table cmpl-syntax-table)
  (condition-case nil
      (cond 
       ;; Cursor is on following-char and after preceding-char
       ((memq (char-syntax (following-char)) '(?w ?_))     
	(setq cmpl-saved-point (point)
	      cmpl-symbol-start (scan-sexps (1+ cmpl-saved-point) -1)
	      cmpl-symbol-end (scan-sexps cmpl-saved-point 1))
	;; remove chars to ignore at the start
	(cond ((= (char-syntax (char-after cmpl-symbol-start)) ?w)
	       (goto-char cmpl-symbol-start)
	       (forward-word 1)
	       (setq cmpl-symbol-start (point))
	       (goto-char cmpl-saved-point)
	       ))
	;; remove chars to ignore at the end
	(cond ((= (char-syntax (char-after (1- cmpl-symbol-end))) ?w)
	       (goto-char cmpl-symbol-end)
	       (forward-word -1)
	       (setq cmpl-symbol-end (point))
	       (goto-char cmpl-saved-point)
	       ))
	;; restore state
	(set-syntax-table cmpl-saved-syntax)
	;; Return completion if the length is reasonable
	(if (and (<= (cmpl-read-time-eval completion-min-length)
		     (- cmpl-symbol-end cmpl-symbol-start))
		 (<= (- cmpl-symbol-end cmpl-symbol-start)
		     (cmpl-read-time-eval completion-max-length)))
	    (buffer-substring cmpl-symbol-start cmpl-symbol-end))
	)
       (t ;; restore table if no symbol
	(set-syntax-table cmpl-saved-syntax)
	nil))
      (error ;; restore table if no symbol
       (set-syntax-table cmpl-saved-syntax)
       nil)))

(defun pvs-symbol-before-point ()
  "Returns a string of the symbol immediately before point.
Returns nil if there isn't one longer than `completion-min-length'."       
  ;; This is called when a word separator is typed so it must be FAST !
  (setq cmpl-saved-syntax (syntax-table))
  (set-syntax-table cmpl-syntax-table)
  (condition-case nil
      ;; Cursor is on following-char and after preceding-char
      (cond
       ((= (setq cmpl-preceding-syntax (char-syntax (preceding-char))) ?_)
	;; No chars. to ignore at end
	(setq cmpl-symbol-end (point)
	      cmpl-symbol-start (scan-sexps (1+ cmpl-symbol-end) -1)
	      )
	;; remove chars to ignore at the start
	(cond ((= (char-syntax (char-after cmpl-symbol-start)) ?w)
	       (goto-char cmpl-symbol-start)
	       (forward-word 1)
	       (setq cmpl-symbol-start (point))
	       (goto-char cmpl-symbol-end)
	       ))
	;; restore state
	(set-syntax-table cmpl-saved-syntax)
	;; return value if long enough
	(if (>= cmpl-symbol-end
		(+ cmpl-symbol-start
		   (cmpl-read-time-eval completion-min-length)))
	    (buffer-substring cmpl-symbol-start cmpl-symbol-end))
	)
       ((= cmpl-preceding-syntax ?w)
	;; chars to ignore at end
	(setq cmpl-saved-point (point)
	      cmpl-symbol-start (scan-sexps (1+ cmpl-saved-point) -1))
	;; take off chars. from end
	(forward-word -1)
	(setq cmpl-symbol-end (point))
	;; remove chars to ignore at the start
	(cond ((= (char-syntax (char-after cmpl-symbol-start)) ?w)
	       (goto-char cmpl-symbol-start)
	       (forward-word 1)
	       (setq cmpl-symbol-start (point))
	       ))
	;; restore state
	(goto-char cmpl-saved-point)
	(set-syntax-table cmpl-saved-syntax)
	;; Return completion if the length is reasonable
	(if (and (<= (cmpl-read-time-eval completion-min-length)
		     (- cmpl-symbol-end cmpl-symbol-start))
		 (<= (- cmpl-symbol-end cmpl-symbol-start)
		     (cmpl-read-time-eval completion-max-length)))
	    (buffer-substring cmpl-symbol-start cmpl-symbol-end))
	)
       (t 
	;; restore table if no symbol
	(set-syntax-table cmpl-saved-syntax)
	nil))
      (error ;; restore table if no symbol
        (set-syntax-table cmpl-saved-syntax)
	nil)))

(defun pvs-symbol-before-point-for-complete ()
  ;; "Returns a string of the symbol immediately before point
  ;; or nil if there isn't one.  Like symbol-before-point but doesn't trim the
  ;; end chars."
  ;; Cursor is on following-char and after preceding-char
  (setq cmpl-saved-syntax (syntax-table))
  (set-syntax-table cmpl-syntax-table)
  (condition-case nil
      (cond ((memq (setq cmpl-preceding-syntax (char-syntax (preceding-char)))
		   '(?_ ?w))
	     (setq cmpl-symbol-end (point)
		   cmpl-symbol-start (scan-sexps (1+ cmpl-symbol-end) -1)
		   )
	     ;; remove chars to ignore at the start
	     (cond ((= (char-syntax (char-after cmpl-symbol-start)) ?w)
		    (goto-char cmpl-symbol-start)
		    (forward-word 1)
		    (setq cmpl-symbol-start (point))
		    (goto-char cmpl-symbol-end)
		    ))
	     ;; restore state
	     (set-syntax-table cmpl-saved-syntax)
	     ;; Return completion if the length is reasonable
	     (if (and (<= (cmpl-read-time-eval
			   completion-prefix-min-length)
			  (- cmpl-symbol-end cmpl-symbol-start))
		      (<= (- cmpl-symbol-end cmpl-symbol-start)
			  (cmpl-read-time-eval completion-max-length)))
		 (buffer-substring cmpl-symbol-start cmpl-symbol-end))
	     )
	    (t 
	     ;; restore table if no symbol
	     (set-syntax-table cmpl-saved-syntax)
	     nil)
	    )
    (error ;; restore table if no symbol
	   (set-syntax-table cmpl-saved-syntax)
	   nil)
    ))

(defun pvs-set-completion-functions ()
  (fset 'symbol-before-point 'pvs-symbol-before-point)
  (fset 'symbol-under-point 'pvs-symbol-under-point)
  (fset 'symbol-before-point-for-complete
	'pvs-symbol-before-point-for-complete)
  nil)

(eval-after-load "completion" '(pvs-set-completion-functions))

(pvs-set-completion-functions)
)
